package main

import (
	"fmt"
	"log"
	mrand "math/rand"
	"net"
	"net/rpc"
	"sync"
	"time"

	. "dcnet"
)

type Server struct {
	numClients       int
	anonymityServers []string
	cleartext        [][]byte
	responses        [][]byte

	serverWG      *sync.WaitGroup
	ctxtChan      chan [][]byte
	serverRegChan chan serverRegParams
	clientRegChan chan *chan ClientRegistration

	waitResponses chan bool
	serverRegDone chan bool
	clientRegDone chan bool
}

func NewServer() *Server {
	s := Server{
		numClients:       0,
		anonymityServers: make([]string, 0, 32),

		serverWG:      new(sync.WaitGroup),
		ctxtChan:      make(chan [][]byte),
		serverRegChan: make(chan serverRegParams),
		clientRegChan: make(chan *chan ClientRegistration),

		waitResponses: make(chan bool),
		clientRegDone: make(chan bool),
		serverRegDone: make(chan bool),
	}

	return &s
}

func (s *Server) GetResponses(_ int, r *[][]byte) error {
	<-s.waitResponses

	*r = s.responses
	return nil
}

func (s *Server) GetCleartext(_ int, ctxt *[][]byte) error {
	s.serverWG.Wait()
	*ctxt = s.cleartext
	return nil
}

func (s *Server) AddServerCiphertext(ctxt [][]byte, _ *int) error {
	s.ctxtChan <- ctxt
	return nil
}

func (s *Server) gatherServerCiphertexts() {
	s.cleartext = make([][]byte, s.numClients)
	for i := range s.cleartext {
		s.cleartext[i] = make([]byte, MsgLen)
	}

	for m := 0; m < len(s.anonymityServers); m++ {
		ctxt := <-s.ctxtChan

		for i, ctxt := range ctxt {
			Xor(ctxt, s.cleartext[i])
		}

		s.serverWG.Done()
	}
}

func (s *Server) RegisterClient(_ int, r *ClientRegistration) error {
	clientChan := make(chan ClientRegistration)
	s.clientRegChan <- &clientChan
	*r = <-clientChan
	return nil
}

type serverRegParams struct {
	serverName string
	serverChan chan ServerRegistration
}

func (s *Server) RegisterServer(serverName string, r *ServerRegistration) error {
	p := serverRegParams{serverName, make(chan ServerRegistration)}
	s.serverRegChan <- p
	*r = <-p.serverChan
	return nil
}

func (s *Server) clientRegistrationLoop() {
	chans := make([]*chan ClientRegistration, 0, 1024)

	for {
		clientChan := <-s.clientRegChan
		s.numClients++
		chans = append(chans, clientChan)

		// TODO we may want to add a grace period here
		// TODO could probably do this more elegantly with a WaitGroup
		if s.numClients >= MinClients {
			// TODO can't add more clients while waiting for servers
			// and vice versa in the server registration loop
			//fmt.Println("*** got enough clients:", s.numClients)
			s.clientRegDone <- true
			<-s.serverRegDone

			pi := mrand.Perm(s.numClients)
			for i, ch := range chans {
				r := ClientRegistration{pi[i], s.numClients, s.anonymityServers}
				*ch <- r
			}

			return
		}
	}
}

func (s *Server) serverRegistrationLoop() {
	chans := make([]chan ServerRegistration, 0, 32)

	for {
		p := <-s.serverRegChan
		s.anonymityServers = append(s.anonymityServers, p.serverName)
		chans = append(chans, p.serverChan)
		s.serverWG.Add(1)

		// TODO we may want to add a grace period here
		if len(s.anonymityServers) >= MinServers {
			//fmt.Println("*** got enough servers:", s.anonymityServers)
			<-s.clientRegDone
			s.serverRegDone <- true

			for _, ch := range chans {
				ch <- ServerRegistration{s.numClients}
			}

			return
		}
	}
}

func (s *Server) readMessages() []string {
	messages := make([]string, len(s.cleartext))

	for i, x := range s.cleartext {
		messages[i] = CString(x)
	}

	return messages
}

func (s *Server) round() {
	go s.clientRegistrationLoop()
	s.serverRegistrationLoop()

	s.gatherServerCiphertexts()
	messages := s.cleartext //readMessages()

	responses := FetchReqs(messages)
	s.responses = responses
	close(s.waitResponses)
}

func (s *Server) cleanup() {
	s.numClients = 0
	s.anonymityServers = s.anonymityServers[:0]
	s.waitResponses = make(chan bool)
}

func (s *Server) serverLoop() {
	for i := 1; ; i++ {
		fmt.Println("[", i, "]")
		s.round()

		time.Sleep(500 * time.Millisecond)
		s.cleanup()
	}
}

func main() {
	mrand.Seed(time.Now().UnixNano())

	s := NewServer()
	server := rpc.NewServer()
	server.Register(s)

	go s.serverLoop()

	l, e := net.Listen("tcp", ":61111")
	if e != nil {
		log.Fatal("listen error: ", e)
	}
	fmt.Println("control server listening...")
	server.Accept(l)
}
