package main

import (
	"bytes"
	"flag"
	"io"
	"log"
	"net"
	"net/http"
	"net/rpc"
	"sync"
	"syscall"

	. "dcnet"
)

type ProxyRPC struct {
	req    []byte
	resp   []byte
	finish sync.Mutex //program finish
	rg     sync.Mutex //synchronize order of operation
	wg     sync.Mutex //synchronize order of operation
}

type ProxyServer struct {
	client   *http.Client
	proxyRPC *ProxyRPC
}

//func (p *ProxyRPC) GetRequest(_ int, r *http.Request) error {
func (p *ProxyRPC) GetRequest(_ int, r *[]byte) error {
	p.rg.Lock()
	*r = p.req
	return nil
}

func (p *ProxyRPC) PutResponse(resp []byte, _ *int) error {
	p.resp = resp
	p.wg.Unlock()
	return nil
}

func CopyHeaders(dst http.Header, src http.Header) {
	for k, _ := range dst {
		dst.Del(k)
	}
	for k, vs := range src {
		for _, v := range vs {
			dst.Add(k, v)
		}
	}
}

func (p *ProxyServer) ServeHTTP(w http.ResponseWriter, r *http.Request) {
	client := &http.Client{}
	if r.Method == "CONNECT" {
		return
	}

	req, err := http.NewRequest(r.Method, r.RequestURI, r.Body)
	if err != nil {
		log.Fatal("Couldn't generate request: ", err)
	}
	//p.proxyRPC.req = req

	//p.proxyRPC.rg.Unlock()

	//p.proxyRPC.wg.Lock()
	resp, _ := client.Do(req)
	//resp := p.proxyRPC.resp
	CopyHeaders(w.Header(), resp.Header)
	w.WriteHeader(resp.StatusCode)
	_, err = io.Copy(w, resp.Body)
	if err != nil && err != syscall.EPIPE {
		log.Fatal("Couldn't copy response body: ", err)
	}
	err = resp.Body.Close()
	if err != nil {
		log.Fatal("Can't close response body: ", err)
	}
}

func NewProxyRPC() *ProxyRPC {
	return &ProxyRPC{}
}

func NewProxyServer() *ProxyServer {
	return &ProxyServer{
		client:   &http.Client{},
		proxyRPC: NewProxyRPC(),
	}
}

func RawServer(p *ProxyServer) {
	l, err := net.Listen("tcp", ":63330")
	if err != nil {
		log.Fatal("listen error:", err)
	}

	for {
		conn, _ := l.Accept()
		msg := make([]byte, MsgLen)

		n, err := conn.Read(msg)
		if n == 0 {
			continue
		}
		if err != nil {
			log.Fatal("Can't read request: ", err)
		}

		println(string(msg))

		p.proxyRPC.req = msg

		p.proxyRPC.rg.Unlock()

		p.proxyRPC.wg.Lock()

		n, _ = conn.Write(bytes.TrimRight(p.proxyRPC.resp, "\x00"))
		conn.Close()
	}
}

func StartHTTPServer(p *ProxyServer) {
	RawServer(p)
	log.Fatal(http.ListenAndServe(":63330", p))
}

func StartRPCServer(p *ProxyServer) {
	var addr *string = flag.String("l", ":43330", "listen addr:port")
	flag.Parse()

	server := rpc.NewServer()
	server.Register(p.proxyRPC)
	l, err := net.Listen("tcp", *addr)
	if err != nil {
		log.Fatal("listen error:", err)
	}
	server.Accept(l)
}

func main() {
	p := NewProxyServer()

	p.proxyRPC.rg.Lock()
	p.proxyRPC.wg.Lock()
	p.proxyRPC.finish.Lock()

	go StartHTTPServer(p)
	go StartRPCServer(p)

	p.proxyRPC.finish.Lock()
	p.proxyRPC.finish.Unlock()
}
